/*
 Copyright (C) MINZKN.COM
 All rights reserved.
 Code by JaeHyuk Cho <mailto:minzkn@minzkn.com>
*/

/*
#if !defined(_GNU_SOURCE)
# define _GNU_SOURCE
#endif
*/

#include <sys/types.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <netinet/ip.h>
#include <netinet/udp.h>

#define __hwport_inline __inline

typedef uint16_t hwport_sum16_t;
#define hwport_csum_mangled_0 ((hwport_sum16_t)0xffffu) /* -0 */
typedef uint32_t hwport_wsum_t;
#define hwport_wsum_carry(m_sum) ((m_sum) >> 16)
#define hwport_wsum_carry_mask(m_sum) ((m_sum) & ((hwport_wsum_t)0xffffu))

hwport_wsum_t hwport_do_csum(const void *data, size_t size);

static __hwport_inline uint16_t hwport_bswap_16(uint16_t value)
{
	return ((value >> 8) & ((uint16_t)0xffu)) | ((value & ((uint16_t)0xffu)) << 8);
}

static __hwport_inline hwport_sum16_t hwport_wcsum_to_sum16(hwport_wsum_t sum)
{
	sum = hwport_wsum_carry_mask(sum) + hwport_wsum_carry(sum);
	return (hwport_sum16_t)(hwport_wsum_carry_mask(sum) + hwport_wsum_carry(sum));
}

static __hwport_inline hwport_sum16_t hwport_csum_fold(hwport_wsum_t sum)
{
	return (hwport_sum16_t)(~hwport_wcsum_to_sum16(sum));
}

static __hwport_inline hwport_wsum_t hwport_csum_unfold(hwport_sum16_t sum)
{
	return (hwport_wsum_t)sum;
}

static __hwport_inline hwport_wsum_t hwport_csum_add(hwport_wsum_t sum, hwport_wsum_t addend)
{
	sum += addend;
	return sum + (sum < addend);
}

static __hwport_inline hwport_sum16_t hwport_csum16_add(hwport_sum16_t sum, hwport_sum16_t addend)
{
	sum += addend;
	return sum + (sum < addend);
}

static __hwport_inline hwport_wsum_t hwport_csum_sub(hwport_wsum_t sum, hwport_wsum_t addend)
{
	return hwport_csum_add(sum, ~addend);
}

static __hwport_inline hwport_sum16_t hwport_csum16_sub(hwport_sum16_t sum, hwport_sum16_t addend)
{
	return hwport_csum16_add(sum, ~addend);
}

static __hwport_inline void hwport_csum_replace(hwport_wsum_t *sum_ptr, hwport_wsum_t from, hwport_wsum_t to)
{
	*sum_ptr = ~hwport_csum_add(hwport_csum_sub(*sum_ptr, from), to);
}

static __hwport_inline void hwport_csum_replace4(hwport_sum16_t *sum_ptr, uint32_t from, uint32_t to)
{
	*sum_ptr = hwport_csum_fold(hwport_csum_add(hwport_csum_sub(~hwport_csum_unfold(*sum_ptr), (hwport_wsum_t)from), (hwport_wsum_t)to));
}

static __hwport_inline void hwport_csum_replace2(hwport_sum16_t *sum_ptr, uint16_t from, uint16_t to)
{
	*sum_ptr = ~hwport_csum16_add(hwport_csum16_sub(~(*sum_ptr), (hwport_sum16_t)from), (hwport_sum16_t)to);
}

static __hwport_inline int hwport_ip_decrease_ttl(struct iphdr *iph)
{
	hwport_wsum_t sum = (hwport_wsum_t)iph->check;

	sum += (hwport_wsum_t)htons(0x0100);
	iph->check = (hwport_sum16_t)(sum + (sum >= ((hwport_wsum_t)0xffffu)));

	return --iph->ttl;
}

static __hwport_inline hwport_wsum_t hwport_csum_partial(const void *data, size_t size, hwport_wsum_t sum)
{
	return hwport_csum_add(hwport_do_csum(data, size), sum);
}

static __hwport_inline hwport_sum16_t hwport_ip_compute_csum(const void *data, size_t size)
{
	return hwport_csum_fold(hwport_do_csum(data, size));
}

static __hwport_inline hwport_sum16_t hwport_ip_fast_csum(const void *iph, unsigned int ihl)
{ /* hwport_ip_compute_csum optimized, 4 octet boundaries */
#if 0L /* fast, no carry */
	hwport_wsum_t sum = (hwport_wsum_t)0u;

	while (ihl > 0u) {
		sum = hwport_csum_add(sum, (hwport_wsum_t)(*((const uint32_t *)iph)));
		--ihl;
		iph = (const void *)(((const uint8_t *)iph) + sizeof(uint32_t));
	}
	
	return (hwport_sum16_t)(~hwport_wcsum_to_sum16((uint32_t)sum));
#else /* full */
	return hwport_csum_fold(hwport_do_csum(iph, (size_t)(ihl << 2)));
#endif
}

static __hwport_inline void hwport_ip_send_check(struct iphdr *iph)
{
	iph->check = 0;
	iph->check = hwport_ip_fast_csum((const void *)iph, (unsigned int)iph->ihl);
}

static __hwport_inline hwport_wsum_t hwport_csum_tcpudp_nofold(uint32_t saddr, uint32_t daddr, uint32_t len, uint8_t proto, hwport_wsum_t sum)
{
	sum = hwport_csum_add(sum, (hwport_wsum_t)saddr);
	sum = hwport_csum_add(sum, (hwport_wsum_t)daddr);
#if defined(__LITTLE_ENDIAN)
	sum = hwport_csum_add(sum, (((hwport_wsum_t)proto) + ((hwport_wsum_t)len)) << 8);
#elif defined(__BIG_ENDIAN)
	sum = hwport_csum_add(sum, ((hwport_wsum_t)proto) + ((hwport_wsum_t)len));
#else
	sum = hwport_csum_add(sum, (hwport_wsum_t)htonl(len));
	sum = hwport_csum_add(sum, (hwport_wsum_t)htonl(proto));
#endif

	return sum;
}

static __hwport_inline hwport_sum16_t hwport_csum_tcpudp_magic(uint32_t saddr, uint32_t daddr, uint32_t len, uint8_t proto, hwport_wsum_t sum)
{
	return hwport_csum_fold(hwport_csum_tcpudp_nofold(saddr, daddr, len, proto, sum));
}

static __hwport_inline hwport_sum16_t hwport_csum_ipv6_magic(const struct in6_addr *saddr, const struct in6_addr *daddr, uint32_t len, uint8_t proto, hwport_wsum_t sum)
{
#if 0L && defined(__USE_MISC)
	sum = hwport_csum_add(sum, (hwport_wsum_t)saddr->s6_addr32[0]);
	sum = hwport_csum_add(sum, (hwport_wsum_t)saddr->s6_addr32[1]);
	sum = hwport_csum_add(sum, (hwport_wsum_t)saddr->s6_addr32[2]);
	sum = hwport_csum_add(sum, (hwport_wsum_t)saddr->s6_addr32[3]);

	sum = hwport_csum_add(sum, (hwport_wsum_t)daddr->s6_addr32[0]);
	sum = hwport_csum_add(sum, (hwport_wsum_t)daddr->s6_addr32[1]);
	sum = hwport_csum_add(sum, (hwport_wsum_t)daddr->s6_addr32[2]);
	sum = hwport_csum_add(sum, (hwport_wsum_t)daddr->s6_addr32[3]);
#elif 0L
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&saddr->s6_addr)) + 0));
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&saddr->s6_addr)) + 1));
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&saddr->s6_addr)) + 2));
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&saddr->s6_addr)) + 3));

	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&daddr->s6_addr)) + 0));
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&daddr->s6_addr)) + 1));
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&daddr->s6_addr)) + 2));
	sum = hwport_csum_add(sum, (hwport_wsum_t)*(((const uint32_t *)(&daddr->s6_addr)) + 3));
#else
	sum = hwport_csum_partial((const void *)saddr, sizeof(*saddr), sum);
	sum = hwport_csum_partial((const void *)daddr, sizeof(*saddr), sum);
#endif
	sum = hwport_csum_add(sum, (hwport_wsum_t)htonl(len));
	sum = hwport_csum_add(sum, (hwport_wsum_t)htonl(proto));

	return hwport_csum_fold(sum);
}

static __hwport_inline hwport_sum16_t hwport_tcp_v4_check(uint32_t len, uint32_t saddr, uint32_t daddr, hwport_wsum_t sum)
{
	return hwport_csum_tcpudp_magic(saddr, daddr, len, IPPROTO_TCP, sum);
}

static __hwport_inline hwport_sum16_t hwport_tcp_v6_check(uint32_t len, const struct in6_addr *saddr, const struct in6_addr *daddr, hwport_wsum_t sum)
{
	return hwport_csum_ipv6_magic(saddr, daddr, len, IPPROTO_TCP, sum);
}

static __hwport_inline hwport_sum16_t hwport_udp_v4_check(uint32_t len, uint32_t saddr, uint32_t daddr, hwport_wsum_t sum)
{
	return hwport_csum_tcpudp_magic(saddr, daddr, len, IPPROTO_UDP, sum);
}

static __hwport_inline hwport_sum16_t hwport_udp_v6_check(uint32_t len, const struct in6_addr *saddr, const struct in6_addr *daddr, hwport_wsum_t sum)
{
	return hwport_csum_ipv6_magic(saddr, daddr, len, IPPROTO_UDP, sum);
}

int main(int s_argc, char **s_argv);

hwport_wsum_t hwport_do_csum(const void *data, size_t size)
{
	hwport_wsum_t sum = (hwport_wsum_t)0u;
	int is_odd;

	if (size <= ((size_t)0u)) {
		return sum;
	}

	is_odd = (int)(1ul & ((unsigned long)data)) ;
	if (is_odd != 0) {
#if defined(__LITTLE_ENDIAN)
		sum = ((hwport_wsum_t)(*((const uint8_t *)data))) << 8;
#elif defined(__BIG_ENDIAN)
		sum = (hwport_wsum_t)(*((const uint8_t *)data));
#else
                sum = (hwport_wsum_t)htons((uint16_t)*((const uint8_t *)data));
#endif
		--size;
		data = (const void *)(((const uint8_t *)data) + ((size_t)1u));
	}

	while (size >= sizeof(hwport_sum16_t)) {
		sum = hwport_csum_add(sum, (hwport_wsum_t)(*((const uint16_t *)data)));
		size -= sizeof(uint16_t);
		data = (const void *)(((const uint8_t *)data) + sizeof(uint16_t));
	}
	
	if (size > ((size_t)0u)) {
#if defined(__LITTLE_ENDIAN)
		sum = hwport_csum_add(sum, (hwport_wsum_t)(*((const uint8_t *)data)));
#elif defined(__BIG_ENDIAN)
		sum = hwport_csum_add(sum, ((hwport_wsum_t)(*((const uint8_t *)data))) << 8);
#else
		sum = hwport_csum_add(sum, (hwport_wsum_t)htons(((uint16_t)(*((const uint8_t *)data))) << 8));
#endif
	}

	if (is_odd != 0) {
		sum = (hwport_wsum_t)hwport_bswap_16(hwport_wcsum_to_sum16(sum));
	}

	return sum;
}

int main(int s_argc, char **s_argv)
{
	(void)s_argc;
	(void)s_argv;
	(void)fprintf(stdout, "BEGIN: %s\n", s_argv[0]);

	do {
		size_t buffer_size;
		void *buffer;

		void *head, *end, *network_header, *transport_header, *payload;
		size_t truesize, mtu, len, offset, payload_size;

		struct iphdr *iph;
		struct udphdr *uh;

		hwport_sum16_t iph_check1, iph_check2;
		hwport_sum16_t uh_check1, uh_check2;
		hwport_sum16_t verify_iph_check1, verify_iph_check2;
		hwport_sum16_t verify_uh_check1, verify_uh_check2;

		uint32_t temp32;
		uint16_t temp16;

		buffer_size = ((size_t)128u) << 10;
		buffer = malloc(buffer_size);
		if (buffer == NULL) {
			(void)fprintf(stderr, "ERROR: buffer alloc failed ! (buffer_size=%lu)\n", (unsigned long)buffer_size);
			return EXIT_FAILURE;
		}

		head = (void *)(((uint8_t *)buffer) + 0);
		end = (void *)(((uint8_t *)buffer) + buffer_size);
		truesize = (size_t)(end - head);
		if (buffer_size != truesize) {
			(void)fprintf(stderr, "ERROR: buffer_size != truesize ! (buffer_size=%lu, truesize=%lu)\n", (unsigned long)buffer_size, (unsigned long)truesize);
		}
		(void)memset(head, 0, truesize);

		mtu = (size_t)1500u;
		len = (size_t)0u;

		/* reserve buffer */
#if 0L /* not aligned (odd) pointer test */
		offset = (size_t)(16u + 1u);
#else
		offset = (size_t)16u;
#endif

		/* put L3 header pointing */
		iph = network_header = (void *)(((uint8_t *)head) + offset);
		len += sizeof(*iph);
		
		/* put L4 header pointing */
		uh = transport_header = (void *)(&iph[1]);
		len += sizeof(*uh);
		
		/* put L4 payload pointing */
		payload_size = mtu - (sizeof(*iph) + sizeof(*uh));
		payload = (void *)(&uh[1]);
		len += payload_size;

		/* build L3 header */
		iph->version = 4;
		iph->ihl = sizeof(*iph) >> 2; /* 4 boundary */
		iph->tos = 0u;
		iph->tot_len = htons(len);
		iph->id = htons(0x1234u);
		iph->frag_off = htons(0x0000u);
		iph->ttl = 64;
		iph->protocol = IPPROTO_UDP;
		iph->check = 0u; /* initial L3 checksum */
		iph->saddr = htonl(0x01020304); 
		iph->daddr = htonl(0x01010101); 

		/* build L4 header */
		uh->source = htons(1024u);
		uh->dest = htons(4500u);
		uh->len = htons(sizeof(*uh) + payload_size);
		uh->check = 0u; /* initial L4 checksum */

		/* build L4 payload */
		for (offset = 0u;offset < payload_size;offset++) {
			*(((uint8_t *)payload) + offset) = (uint8_t)offset;
		}
#if 0L /* make UDP checksum mangled for TEST */
		do {
			int is_mangled = 0;

			while (is_mangled == 0) {
				for (offset = 0u;offset < payload_size;offset++) {
					if(hwport_csum_tcpudp_magic(iph->saddr, iph->daddr, (uint32_t)(iph->ihl << 2), iph->protocol,
						hwport_csum_partial((const void *)uh, (size_t)ntohs(uh->len), (hwport_wsum_t)0u)
					  ) == ((hwport_sum16_t)0u)) {
						is_mangled = 1;
						break;
					}
					if ((*(((uint8_t *)payload) + offset)) < ((uint8_t)0xffu)) {
						++(*(((uint8_t *)payload) + offset));
					}
				}
			}
		} while(0);
#endif

		/* do ip checksum */
		iph_check1 = hwport_csum_fold(hwport_csum_partial((const void *)iph, (size_t)(iph->ihl << 2), (hwport_wsum_t)0u));
		iph_check2 = hwport_ip_compute_csum((const void *)iph, (size_t)(iph->ihl << 2));
#if 0L
		hwport_ip_send_check(iph);
#else
		iph->check = 0u;
		iph->check = hwport_ip_fast_csum((const void *)iph, (unsigned int)iph->ihl);
#endif

		/* do transport checksum */
		uh_check1 = hwport_csum_fold(
			hwport_csum_tcpudp_nofold(iph->saddr, iph->daddr, (uint32_t)(iph->ihl << 2), iph->protocol,
                       		hwport_csum_partial((const void *)uh, (size_t)ntohs(uh->len), (hwport_wsum_t)0u)
			)
               	);
		uh_check2 = hwport_csum_fold(
			hwport_csum_add(
				hwport_csum_tcpudp_nofold(iph->saddr, iph->daddr, (uint32_t)(iph->ihl << 2), iph->protocol, (hwport_wsum_t)0u),
                       		hwport_csum_partial((const void *)uh, (size_t)ntohs(uh->len), (hwport_wsum_t)0u)
			)
		);
		uh->check = hwport_csum_tcpudp_magic(iph->saddr, iph->daddr, (uint32_t)(iph->ihl << 2), iph->protocol,
			hwport_csum_partial((const void *)uh, (size_t)ntohs(uh->len), (hwport_wsum_t)0u)
		);
#if 1L /* UDP mangled 0 if zero checksum */
		if (iph->protocol == IPPROTO_UDP && uh->check == 0u /* +0 */) {
			(void)fprintf(stdout, "UDP MANGLED 0 packet\n");
			uh->check = hwport_csum_mangled_0; /* 0xffff(-0) : UDP MANGLED 0 checksum */
		}
#endif

		/* verify ip checksum */
		verify_iph_check1 = hwport_wcsum_to_sum16(hwport_csum_partial((const void *)iph, (size_t)(iph->ihl << 2), (hwport_wsum_t)0u));
		verify_iph_check2 = hwport_csum_fold(hwport_csum_partial((const void *)iph, (size_t)(iph->ihl << 2), (hwport_wsum_t)0u));

		/* verify transport checksum */
		verify_uh_check1 = hwport_wcsum_to_sum16(
			hwport_csum_tcpudp_nofold(iph->saddr, iph->daddr, (uint32_t)(iph->ihl << 2), iph->protocol,
                        	hwport_csum_partial((const void *)uh, (size_t)ntohs(uh->len), (hwport_wsum_t)0u)
                	)
		);
		verify_uh_check2 = hwport_csum_tcpudp_magic(iph->saddr, iph->daddr, (uint32_t)(iph->ihl << 2), iph->protocol,
                        hwport_csum_partial((const void *)uh, (size_t)ntohs(uh->len), (hwport_wsum_t)0u)
                );

		/* report */
		(void)fprintf(stdout, "\n[IPv4+UDP] iph & uh checksum test\n");
		(void)fprintf(stdout, "\tiph_check1=%04XH, iph_check2=%04XH, iph->check=%04XH\n",
			(unsigned int)iph_check1,
		       	(unsigned int)iph_check2,
		       	(unsigned int)iph->check
		);
		(void)fprintf(stdout, "\tuh_check1=%04XH, uh_check2=%04XH, uh->check=%04XH\n",
			(unsigned int)uh_check1,
		       	(unsigned int)uh_check2,
			(unsigned int)uh->check
		);

		(void)fprintf(stdout, "\n[IPv4+UDP] iph & uh checksum verify test\n");
		(void)fprintf(stdout, "\tverify_iph_check1=%04XH(%s), verify_iph_check2=%04XH(%s)\n",
			(unsigned int)verify_iph_check1, (verify_iph_check1 == ((hwport_sum16_t)0xffffu)) ? "PASSED" : "FAILED",
			(unsigned int)verify_iph_check2, (verify_iph_check2 == ((hwport_sum16_t)0x0000u)) ? "PASSED" : "FAILED"
		);
		(void)fprintf(stdout, "\tverify_uh_check1=%04XH(%s), verify_uh_check2=%04XH(%s)\n",
			(unsigned int)verify_uh_check1, (verify_uh_check1 == ((hwport_sum16_t)0xffffu)) ? "PASSED" : "FAILED",
		       	(unsigned int)verify_uh_check2, (verify_uh_check2 == ((hwport_sum16_t)0x0000u)) ? "PASSED" : "FAILED"
		);

		(void)fprintf(stdout, "\n[IPv4+UDP+NAT] iph->daddr change & checksum update test\n");
		temp32 = iph->daddr;
		iph->daddr++; /* address translation */
		hwport_csum_replace4((hwport_sum16_t *)(&iph->check), temp32, (uint32_t)iph->daddr);
		hwport_csum_replace4((hwport_sum16_t *)(&uh->check), temp32, (uint32_t)iph->daddr);
		(void)fprintf(stdout, "\tiph->daddr=%08XH to %08XH, iph->check=%04XH, uh->check=%04XH\n",
			(unsigned int)ntohl(temp32),
			(unsigned int)ntohl(iph->daddr),
		       	(unsigned int)iph->check,
		       	(unsigned int)uh->check
		);
		
		(void)fprintf(stdout, "\n[IPv4+UDP+PAT] uh->dest change & checksum update test\n");
		temp16 = uh->dest;
		uh->dest = htons(ntohs(uh->dest) + 1);
		hwport_csum_replace2((hwport_sum16_t *)(&uh->check), temp16, (uint16_t)uh->dest);
		(void)fprintf(stdout, "\tuh->dest=%u to %u, uh->check=%04XH\n",
			(unsigned int)ntohs(temp16),
			(unsigned int)ntohs(uh->dest),
		       	(unsigned int)uh->check
		);
		
		(void)fprintf(stdout, "\n[1's complement] -0 & +0 checksum test (16-bits : A + B == FFFFH[csum=0000H], A + B + 0000H == A + B + FFFFH )\n");
		temp16 = (uint16_t)0x1234u;
		temp32 = ((uint32_t)temp16) | (((uint32_t)(~temp16)) << 16); /* input for mangled checksum */
		temp16 = hwport_csum_fold(hwport_csum_partial((const void *)(&temp32), sizeof(temp32), (hwport_wsum_t)0u));
		(void)fprintf(stdout, "\ttemp32=%08XH (input)\n", (unsigned int)temp32);
		(void)fprintf(stdout, "\tchecksum=%04XH (input checksum, %s)\n", (unsigned int)temp16, (temp16 == ((uint16_t)0u)) ? "+0" : "?ERROR?");
		temp16 = hwport_csum_fold(hwport_csum_partial((const void *)(&temp32), sizeof(temp32), (hwport_wsum_t)hwport_csum_unfold(temp16)));
		(void)fprintf(stdout, "\tchecksum=%04XH (verify input&checksum, %s)\n", (unsigned int)temp16, (temp16 == ((uint16_t)0u)) ? "+0" : "?ERROR?");
		temp16 = hwport_csum_fold(hwport_csum_partial((const void *)(&temp32), sizeof(temp32), (hwport_wsum_t)hwport_csum_mangled_0));
		(void)fprintf(stdout, "\tchecksum=%04XH (verify input&checksum{mangled0[-0, 0xffff]}, %s)\n", (unsigned int)temp16, (temp16 == ((uint16_t)0u)) ? "+0" : "?ERROR?");

		(void)fprintf(stdout, "\n");
		free(buffer);
	} while(0);

	(void)fprintf(stdout, "END  : %s\n", s_argv[0]);

	return(EXIT_SUCCESS);
}

/* End of source */
